# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ GENERAL IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from __future__ import annotations

from typing import TYPE_CHECKING

# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ PROJECT IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from aiva.core.exceptions import DuplicateChatbotKeyError, UnsupportedChatbotError

if TYPE_CHECKING:
    from aiva.core.utils.classes.chatbot import Chatbot


# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ CHATBOT CLASSES
# └─────────────────────────────────────────────────────────────────────────────────────


class ChatbotClasses:
    """A utility class that represents a collection of Chatbot classes"""

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CLASS ATTRIBUTES
    # └─────────────────────────────────────────────────────────────────────────────────

    # Declare type of classes by key
    _classes_by_key: dict[str, type[Chatbot]]

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ __INIT__
    # └─────────────────────────────────────────────────────────────────────────────────

    def __init__(self) -> None:
        """Init Method"""

        # Initialize classes by key
        self._classes_by_key = {}

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ _NORMALIZE KEY
    # └─────────────────────────────────────────────────────────────────────────────────

    def _normalize_key(self, key: str) -> str:
        """Normalizes a Chatbot class key"""

        # Strip key
        key = key.strip()

        # Lowercase key
        key = key.lower()

        # Remove non-alphanumeric characters from key
        key = "".join([char for char in key if char.isalnum()])

        # Return normalized key
        return key

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ INITIALIZE
    # └─────────────────────────────────────────────────────────────────────────────────

    def initialize(self, key: str, api_key: str) -> Chatbot:
        """Initialize a Chatbot instance given a key"""

        # Normalize key
        key_normalized = self._normalize_key(key)

        # Check if key does not exist
        if key_normalized not in self._classes_by_key:
            # Raise UnsupportedChatbotError
            raise UnsupportedChatbotError(f"The chatbot '{key}' is not supported.")

        # Get Chatbot class
        ChatbotClass = self._classes_by_key[key_normalized]

        # Initialize Chatbot instance
        chatbot = ChatbotClass(api_key)

        # Return Chatbot instance
        return chatbot

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ REGISTER
    # └─────────────────────────────────────────────────────────────────────────────────

    def register(self, ChatbotClass: type[Chatbot]) -> None:
        """Registers a Chatbot class"""

        # Get KEYS
        KEYS = ChatbotClass.KEYS

        # Initialize normalized keys
        keys_normalized: set[str] = set()

        # Iterate over keys
        for key in KEYS:
            # Normalize key
            key_normalized = self._normalize_key(key)

            # Check if Chatbot class is already registered
            if key_normalized in self._classes_by_key:
                # Ignore if Chatbot class is the same
                if self._classes_by_key[key_normalized] is ChatbotClass:
                    continue

                # Raise DuplicateChatbotKeyError
                raise DuplicateChatbotKeyError(
                    f"A Chatbot class with the key '{key}' "
                    f"(normalized: {key_normalized}) already exists."
                )

            # Add normalized key to normalized keys
            keys_normalized.add(key_normalized)

        # Add Chatbot class to classes by key
        self._classes_by_key.update({key: ChatbotClass for key in keys_normalized})



# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ GENERAL IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from __future__ import annotations

import logging
import tiktoken

from decimal import Decimal
from openai import ChatCompletion
from typing import Generator, TYPE_CHECKING

# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ PROJECT IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from aiva.core.utils.classes.chatbot import Chatbot
from aiva.core.utils.classes.message import Message
from aiva.core.utils.classes.payload import Payload
from aiva.core.utils.enums.role import Role

if TYPE_CHECKING:
    from aiva.core.types import Kwargs
    from aiva.core.utils.classes.message_instances import MessageInstances


# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ CHAT GPT
# └─────────────────────────────────────────────────────────────────────────────────────


class ChatGPT(Chatbot):
    """An Chatbot class for interfacing with OpenAI's ChatGPT API"""

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CLASS ATTRIBUTES
    # └─────────────────────────────────────────────────────────────────────────────────

    # Define name
    NAME = "ChatGPT"

    # Define keys
    KEYS = ("chat gpt", "chat-gpt", "chat_gpt", "chatgpt", "gpt", "openai")

    # Define model
    MODEL = "gpt-3.5-turbo"

    # Define model encoding
    MODEL_ENCODING = tiktoken.encoding_for_model(MODEL)

    # Define stream
    STREAM = True

    # Define temperature
    TEMPERATURE = 0.8

    # Define token limit
    TOKEN_LIMIT = 4096

    # Define response token info
    TOKENS_FOR_RESPONSE = 500  # Number of tokens available for response
    TOKENS_PER_RESPONSE = 3  # Primed: <|start|>{role/name}<|message|>

    # Define request token info
    TOKENS_PER_REQUEST_MESSAGE = 4  # Primed: <im_start>{role/name}\n{content}<im_end>\n
    TOKENS_PER_REQUEST_MESSAGE_NAME = -1  # Role is omitted if there is a name

    # Define USD cost per token
    COST_USD_PER_TOKEN = Decimal("0.002") / Decimal("1000")

    # See https://github.com/openai/openai-cookbook/
    # blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ COUNT MESSAGE TOKENS
    # └─────────────────────────────────────────────────────────────────────────────────

    def count_message_data_tokens(
        self, message_data: dict[str, str], **kwargs: Kwargs
    ) -> int:
        """Computes and returns the estimated token count of a message data object"""

        # Initialize token count
        token_count = self.TOKENS_PER_REQUEST_MESSAGE

        # Iterate over message items
        for key, text in message_data.items():
            # Increment token count by text token count
            token_count += self.count_text_tokens(text)

            # Check if key is name
            if key == "name":
                # Increment token count by tokens per name
                token_count += self.TOKENS_PER_REQUEST_MESSAGE_NAME

        # Return token count
        return token_count

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ COUNT TEXT TOKENS
    # └─────────────────────────────────────────────────────────────────────────────────

    def count_text_tokens(self, text: str) -> int:
        """Computes and returns the estimated token count of a string"""

        # Encode text
        encoded_text = self.MODEL_ENCODING.encode(text)

        # Get token count
        token_count = len(encoded_text)

        # Return token count of message
        return token_count

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CREATE MESSAGE
    # └─────────────────────────────────────────────────────────────────────────────────

    def create_message(
        self,
        text: str,
        role: Role,
        role_name: str | None = None,
        text_only: bool = False,
        **kwargs: Kwargs,
    ) -> Message:
        """Initializes and returns a Message instance for the API"""

        # Define roles
        roles = {
            Role.ASSISTANT: "assistant",
            Role.SYSTEM: "system",
            Role.USER: "user",
        }

        # Check if role not in roles
        if role not in roles:
            # Emit warning
            logging.warning(f"Role '{role}' not explicitly handled by {self.NAME}.")

            # Set role name
            role_name = role.value

            # Set role
            role = Role.SYSTEM

        # Initialize message data
        data = {
            "role": roles[role],
            **({"name": role_name} if role_name else {}),
            "content": text,
        }

        # Get token count
        token_count = (
            self.count_text_tokens(text)
            if text_only
            else self.count_message_data_tokens(message_data=data)
        )

        # Initialize Message instance
        message = Message(
            text=text,
            role=role,
            role_name=role_name,
            token_count=token_count,
            data=data,
        )

        # Return Message instance
        return message

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CREATE REQUEST PAYLOAD
    # └─────────────────────────────────────────────────────────────────────────────────

    def create_request_payload(
        self,
        prompt: Message,
        message: Message,
        history: MessageInstances,
        stream: bool | None = None,
        temperature: float | None = None,
        **kwargs: Kwargs,
    ) -> Payload:
        """Initializes and returns a Payload instance for the API request"""

        # Get tokens for response
        TOKENS_FOR_RESPONSE = self.TOKENS_FOR_RESPONSE

        # Get tokens per response
        TOKENS_PER_RESPONSE = self.TOKENS_PER_RESPONSE

        # Compute token limit
        token_limit = self.TOKEN_LIMIT - TOKENS_PER_RESPONSE - TOKENS_FOR_RESPONSE

        # Get stream
        stream = stream if stream is not None else self.STREAM

        # Get temperature
        temperature = temperature if temperature is not None else self.TEMPERATURE

        # Initialize payload data
        data = {
            "api_key": self.api_key,
            "model": self.MODEL,
            "temperature": temperature,
            "max_tokens": TOKENS_FOR_RESPONSE,
            "stream": stream,
        }

        # Get token count
        token_count = TOKENS_PER_RESPONSE

        # Initialize Payload instance
        payload = Payload(token_count=token_count, token_limit=token_limit, data=data)

        # Initialize remaining tokens
        remaining_tokens = token_limit - prompt.token_count - message.token_count

        # Get tail of history
        history_tail = history.tail(tokens=remaining_tokens)

        # Add prompt to payload messages
        payload.messages.add(prompt)

        # Add tail of history to payload messages
        payload.messages.add(history_tail)

        # Add message to payload messages
        payload.messages.add(message)

        # Return Payload instance
        return payload

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CREATE RESPONSE PAYLOAD
    # └─────────────────────────────────────────────────────────────────────────────────

    def create_response_payload(self, **kwargs: Kwargs) -> Payload:
        """Initializes and returns a Payload instance for the API response"""

        # Get token count
        token_count = 0

        # Get token limit
        token_limit = self.TOKENS_FOR_RESPONSE

        # Initialize Payload instance
        payload = Payload(token_count=token_count, token_limit=token_limit)

        # Return Payload instance
        return payload

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ SEND
    # └─────────────────────────────────────────────────────────────────────────────────

    def send(
        self, payload: Payload, **kwargs: Kwargs
    ) -> Generator[Message, None, None]:
        """Sends a payload to the API synchronously and yields Message instances"""

        # Make API request
        messages = ChatCompletion.create(**payload.to_json())  # type: ignore

        # Initialize choice key
        message_key = "delta"

        # Check if messages is a dictionary
        if isinstance(messages, dict):
            # Ensure that messages is a list
            messages = [messages]

            # Set choice key to message
            message_key = "message"

        # Initialize is first message to True
        is_first_message = True

        # Iterate over messages
        for message in messages:
            # Get choices
            choices = message["choices"]

            # Get choice
            choice = choices[0]

            # Get choice message
            choice_message = choice[message_key]

            # Continue if content not in choice message
            if "content" not in choice_message:
                continue

            # Extract message text
            message_text = choice_message["content"]

            # Check if is first message
            if is_first_message:
                # Left strip message text
                message_text = message_text.lstrip()

            # Initialize ResponseMessage instance
            response_message = self.create_message(
                text=message_text, role=Role.ASSISTANT, text_only=True
            )

            # Yield ResponseMessage instance
            yield response_message

            # Set is first message to False
            is_first_message = False



# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ GENERAL IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from __future__ import annotations

from typing import Generator, TYPE_CHECKING

# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ PROJECT IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from aiva.core.utils.classes.chatbot import Chatbot
from aiva.core.utils.classes.message import Message
from aiva.core.utils.classes.message_instances import MessageInstances
from aiva.core.utils.classes.transaction import Transaction
from aiva.core.utils.classes.transaction_instances import TransactionInstances

if TYPE_CHECKING:
    from aiva.core.types import Kwargs
    from aiva.core.utils.classes.assistant import Assistant


# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ CHAT
# └─────────────────────────────────────────────────────────────────────────────────────


class Chat:
    """A utility class for managing individual chats threads for a given Chatbot"""

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CLASS ATTRIBUTES
    # └─────────────────────────────────────────────────────────────────────────────────

    # Declare type of assistant
    assistant: Assistant

    # Declare type of bot
    bot: Chatbot

    # Declare type of transactions
    transactions: TransactionInstances

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ __INIT__
    # └─────────────────────────────────────────────────────────────────────────────────

    def __init__(self, assistant: Assistant, bot: Chatbot) -> None:
        """Init Method"""

        # Set assistant
        self.assistant = assistant

        # Set bot
        self.bot = bot

        # Initialize and set transactions
        self.transactions = TransactionInstances()

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ HISTORY
    # └─────────────────────────────────────────────────────────────────────────────────

    @property
    def history(self) -> MessageInstances:
        """Returns a collection of Message instances in the current Chat instance"""

        # Initialize history
        history = MessageInstances()

        # Iterate over transactions
        for transaction in self.transactions:
            # Get request
            request = transaction.request

            # Continue if request is not complete
            if not request.is_complete:
                continue

            # Add request message to history
            history.add(request.message)

            # Get response
            response = transaction.response

            # Check if response is complete
            if response.is_complete:
                # Add response message to history
                history.add(response.message)

        # Return history
        return history

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ SEND
    # └─────────────────────────────────────────────────────────────────────────────────

    def send(self, message: str, **kwargs: Kwargs) -> Generator[Message, None, None]:
        """Sends a synchronous Transaction and yields Message instances"""

        # Initialize Transaction instance
        transaction = Transaction(chat=self, message=message)

        # Add Transaction instance to transactions
        self.transactions.add(transaction)

        # Iterate over Message instances
        yield from transaction.send(**kwargs)



# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ GENERAL IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from __future__ import annotations

from typing import Generator, TYPE_CHECKING

# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ PROJECT IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from aiva.core.exceptions import AlreadySentError
from aiva.core.utils.classes.request import Request
from aiva.core.utils.classes.response import Response

if TYPE_CHECKING:
    from aiva.core.types import Kwargs
    from aiva.core.utils.classes.chat import Chat
    from aiva.core.utils.classes.message import Message


# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ TRANSACTION
# └─────────────────────────────────────────────────────────────────────────────────────


class Transaction:
    """A utility class that normalizes transactions within Chat instances"""

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CLASS ATTRIBUTES
    # └─────────────────────────────────────────────────────────────────────────────────

    # Declare type of chat
    chat: Chat

    # Declare type of request
    request: Request

    # Declare type of response
    response: Response

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ __INIT__
    # └─────────────────────────────────────────────────────────────────────────────────

    def __init__(self, chat: Chat, message: str) -> None:
        """Init Method"""

        # Set chat
        self.chat = chat

        # Initialize and set request
        self.request = Request(transaction=self, message=message)

        # Initialize and set response
        self.response = Response(transaction=self)

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ IS COMPLETE
    # └─────────────────────────────────────────────────────────────────────────────────

    @property
    def is_complete(self) -> bool:
        """Returns a boolean indicating whether the transaction is complete"""

        # Return boolean indicating whether the transaction is complete
        return self.request.is_complete and self.response.is_complete

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ SEND
    # └─────────────────────────────────────────────────────────────────────────────────

    def send(self, **kwargs: Kwargs) -> Generator[Message, None, None]:
        """Sends a synchronous transaction request and yields Message instances"""

        # Check if the transaction is already complete
        if self.is_complete:
            # Raise AlreadySentError
            raise AlreadySentError("This transaction has already been sent.")

        # Send request and yield Message instances
        for message in self.request.send(**kwargs):
            # Mark request as complete
            self.request.is_complete = True

            # Receive and yield message
            yield self.response.receive(message)

        # Mark response as complete
        self.response.is_complete = True



# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ GENERAL IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from __future__ import annotations

from typing import Generator, TYPE_CHECKING

# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ PROJECT IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from aiva.core.utils.classes.chat import Chat
from aiva.core.utils.classes.chat_instances import ChatInstances
from aiva.core.utils.classes.chatbot_classes import ChatbotClasses

if TYPE_CHECKING:
    from aiva.core.types import Kwargs
    from aiva.core.utils.classes.chatbot import Chatbot
    from aiva.core.utils.classes.message import Message


# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ ASSISTANT
# └─────────────────────────────────────────────────────────────────────────────────────


class Assistant:
    """A utility class that represents an chatbot-driven AI assistant"""

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CLASS ATTRIBUTES
    # └─────────────────────────────────────────────────────────────────────────────────

    # Initialize a default assistant prompt
    PROMPT: str = "You are a helpful AI assistant."

    # Initialize Chatbot classes
    CHATBOT_CLASSES: tuple[type[Chatbot], ...] = ()

    # Declare type of this chat
    this_chat: Chat

    # Declare type of all chats
    all_chats: ChatInstances

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ __INIT__
    # └─────────────────────────────────────────────────────────────────────────────────

    def __init__(self, chatbot: str, chatbot_api_key: str = "") -> None:
        """Init Method"""

        # Initialize Chatbot classes
        chatbot_classes = ChatbotClasses()

        # Iterate over Chatbot classes
        for ChatbotClass in self.CHATBOT_CLASSES:
            # Register Chatbot class
            chatbot_classes.register(ChatbotClass=ChatbotClass)

        # Initialize Chatbot instance
        bot = chatbot_classes.initialize(key=chatbot, api_key=chatbot_api_key)

        # Initialize and set this chat
        self.this_chat = Chat(assistant=self, bot=bot)

        # Initialize and set all chats
        self.all_chats = ChatInstances()

        # Add this chat to all chats
        self.all_chats.add(self.this_chat)

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ SEND
    # └─────────────────────────────────────────────────────────────────────────────────

    def send(self, message: str, **kwargs: Kwargs) -> Generator[Message, None, None]:
        """Initiates a Transaction and yields Message instances"""

        # Yield Message instances
        yield from self.this_chat.send(message=message, **kwargs)



# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ GENERAL IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from __future__ import annotations

from typing import TYPE_CHECKING

# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ PROJECT IMPORTS
# └─────────────────────────────────────────────────────────────────────────────────────

from aiva.core.utils.classes.message import Message
from aiva.core.utils.classes.payload import Payload
from aiva.core.utils.enums.role import Role

if TYPE_CHECKING:
    from aiva.core.utils.classes.chat import Chat
    from aiva.core.utils.classes.chatbot import Chatbot
    from aiva.core.utils.classes.transaction import Transaction


# ┌─────────────────────────────────────────────────────────────────────────────────────
# │ RESPONSE
# └─────────────────────────────────────────────────────────────────────────────────────


class Response:
    """A utility class that normalizes transaction responses across chatbot APIs"""

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CLASS ATTRIBUTES
    # └─────────────────────────────────────────────────────────────────────────────────

    # Declare type of transaction
    transaction: Transaction

    # Declare type of payload
    payload: Payload

    # Declare type of is complete
    is_complete: bool

    # Declare type of message
    _message: Message | None

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ __INIT__
    # └─────────────────────────────────────────────────────────────────────────────────

    def __init__(self, transaction: Transaction) -> None:
        """Init Method"""

        # Set transaction
        self.transaction = transaction

        # Initialize and set payload
        self.payload = Payload()

        # Initialize is complete to False
        self.is_complete = False

        # Initialize cached message to None
        self._message = None

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CHAT
    # └─────────────────────────────────────────────────────────────────────────────────

    @property
    def chat(self) -> Chat:
        """Returns the Chat instance associated with the Response instance"""

        # Return the Chat instance
        return self.transaction.chat

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ CHATBOT
    # └─────────────────────────────────────────────────────────────────────────────────

    @property
    def chatbot(self) -> Chatbot:
        """Returns the Chatbot instance associated with the Response instance"""

        # Return the Chatbot instance
        return self.chat.bot

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ MESSAGE
    # └─────────────────────────────────────────────────────────────────────────────────

    @property
    def message(self) -> Message:
        """Returns the Message instance associated with the Response instance"""

        # Get message
        _message = self._message

        # Check if message is None
        if _message is None:
            # Initialize message text
            text = ""

            # Iterate over messages
            for message in self.payload.messages:
                # Update message text
                text += message.text

            # Strip message text
            text = text.strip()

            # Initialize a new Message instance
            _message = self.chatbot.create_message(text=text, role=Role.ASSISTANT)

            # Check if the response is complete
            if self.is_complete:
                # Cache the Message instance
                self._message = _message

        # Return Message instance
        return _message

    # ┌─────────────────────────────────────────────────────────────────────────────────
    # │ RECEIVE
    # └─────────────────────────────────────────────────────────────────────────────────

    def receive(self, message: Message) -> Message:
        """Receives and returns a Message instance"""

        # Add the Message instance to payload messages
        self.payload.messages.add(message)

        # Return Message instance
        return message
